@@ -64,6 +64,7 @@ const Tooltips = imports.ui.tooltips;
 
 const MAX_TEXT_LENGTH = 1000;
 const FLASH_INTERVAL = 500;
+const FLASH_MAX_COUNT = 4;
 
 const WINDOW_PREVIEW_WIDTH = 200;
 const WINDOW_PREVIEW_HEIGHT = 150;
@@ -73,11 +74,12 @@ class WindowPreview extends Tooltips.TooltipBase {
         super(item.actor);
         this._applet = item._applet;
         this.metaWindow = metaWindow;
+        this._windowActor = null;
         let x = this.windowActor;
         this.uiScale = global.ui_scale;
         this.thumbScale = previewScale;
 
-        this._sizeChangedId = null;
+        this._sizeChangedId = 0;
         this.thumbnail = null;
 
         this.actor = new St.BoxLayout({ vertical: true, style_class: "window-list-preview", important: true });
@@ -99,10 +101,14 @@ class WindowPreview extends Tooltips.TooltipBase {
     }
 
     get windowActor() {
-        let actor = this.metaWindow.get_compositor_private();
+        if (this._windowActor) {
+            return this._windowActor;
+        }
 
-        if (actor) {
-            return actor;
+        this._windowActor = this.metaWindow.get_compositor_private();
+
+        if (this._windowActor) {
+            return this._windowActor;
         } else {
             log("metaWindow has no actor!");
             return null;
@@ -173,9 +179,9 @@ class WindowPreview extends Tooltips.TooltipBase {
     }
 
     hide() {
-        if (this._sizeChangedId != null) {
+        if (this._sizeChangedId > 0) {
             this.windowActor.disconnect(this._sizeChangedId);
-            this._sizeChangedId = null;
+            this._sizeChangedId = 0;
         }
         if (this.thumbnail) {
             this.thumbnailBin.set_child(null);
@@ -230,9 +236,9 @@ class WindowPreview extends Tooltips.TooltipBase {
     }
 
     _destroy() {
-        if (this._sizeChangedId != null) {
-            this.windowActor.disconnect(this._sizeChangedId);
-            this.sizeChangedId = null;
+        if (this._sizeChangedId > 0) {
+            this._windowActor.disconnect(this._sizeChangedId);
+            this.sizeChangedId = 0;
         }
         if (this.thumbnail) {
             this.thumbnailBin.set_child(null);
@@ -240,6 +246,7 @@ class WindowPreview extends Tooltips.TooltipBase {
             this.thumbnail = null;
         }
         if (this.actor) {
+            Main.uiGroup.remove_actor(this.actor);
             this.actor.destroy();
             this.actor = null;
         }
@@ -260,6 +267,7 @@ class AppMenuButton {
         this.alert = alert;
         this.labelVisible = false;
         this._signals = new SignalManager.SignalManager();
+        this.xid = metaWindow.get_xwindow();
 
         if (this._applet.orientation == St.Side.TOP)
             this.actor.add_style_class_name('top');
@@ -782,9 +790,8 @@ class AppMenuButton {
                 this.actor.remove_style_class_name(sc);
             else
                 this.actor.add_style_class_name(sc);
-            let result = counter < 4;
-            counter++;
-            return result;
+
+            return counter++ < FLASH_MAX_COUNT;
         });
     }
 };
@@ -939,6 +946,7 @@ class AppMenuButtonRightClickMenu extends Applet.AppletPopupMenu {
             });
         }
         this.addMenuItem(item);
+        item.setSensitive(mw.can_maximize())
 
         item = new PopupMenu.PopupIconMenuItem(_("Close"), "edit-delete", St.IconType.SYMBOLIC);
         this._signals.connect(item, 'activate', function() {
@@ -1012,6 +1020,7 @@ class CinnamonWindowListApplet extends Applet.Applet {
         this.settings.bind("window-preview", "usePreview", this._onPreviewChanged);
         this.settings.bind("window-preview-show-label", "showLabel", this._onPreviewChanged);
         this.settings.bind("window-preview-scale", "previewScale", this._onPreviewChanged);
+        this.settings.bind("last-window-order", "lastWindowOrder", null);
 
         this.signals.connect(global.screen, 'window-added', this._onWindowAddedAsync, this);
         this.signals.connect(global.screen, 'window-monitor-changed', this._onWindowMonitorChanged, this);
@@ -1257,6 +1266,7 @@ class CinnamonWindowListApplet extends Applet.Applet {
             }
         }
 
+        this.refreshing = true;
 
         for (let window of windows) {
             if (this._shouldAdd(window))
@@ -1264,6 +1274,10 @@ class CinnamonWindowListApplet extends Applet.Applet {
             else
                 this._removeWindow(window);
         }
+
+        this.refreshing = false;
+
+        this._applySavedOrder();
     }
 
     _addWindow(metaWindow, alert) {
@@ -1291,6 +1305,8 @@ class CinnamonWindowListApplet extends Applet.Applet {
                 }
             }
         }
+
+        this._saveOrder();
     }
 
     _removeWindow(metaWindow) {
@@ -1302,6 +1318,8 @@ class CinnamonWindowListApplet extends Applet.Applet {
                 this._windows.splice(i, 1);
             }
         }
+
+        this._saveOrder();
     }
 
     _shouldAdd(metaWindow) {
@@ -1310,6 +1328,52 @@ class CinnamonWindowListApplet extends Applet.Applet {
             this._monitorWatchList.indexOf(metaWindow.get_monitor()) != -1;
     }
 
+    /* Store by Windows (XIDs), a simple list
+       xid::xid::xid::xid::xid
+    */
+
+    _applySavedOrder() {
+        let order = this.lastWindowOrder.split("::");
+
+        order.reverse();
+
+        for (let i = 0; i < order.length; i++) {
+            let xid = parseInt(order[i]);
+
+            if (xid === NaN) {
+                continue;
+            }
+
+            let found = this._windows.find(win => (win.xid == xid));
+
+            if (found) {
+                this.manager_container.set_child_at_index(found.actor, 0);
+            }
+        }
+
+        this._saveOrder();
+    }
+
+    _saveOrder() {
+        if (this.refreshing) {
+            return;
+        }
+
+        let new_order = [];
+        let actors = this.manager_container.get_children();
+
+        for (let i = 0; i < actors.length; i++) {
+            new_order.push(actors[i]._delegate.xid);
+        }
+
+        if (new_order.length === 0) {
+            this.lastWindowOrder = "";
+            return;
+        }
+
+        this.lastWindowOrder = new_order.join("::");
+    }
+
     handleDragOver(source, actor, x, y, time) {
         if (this._inEditMode)
             return DND.DragMotionResult.MOVE_DROP;
@@ -1318,20 +1382,17 @@ class CinnamonWindowListApplet extends Applet.Applet {
 
         let children = this.manager_container.get_children();
         let isVertical = this.manager_container.height > this.manager_container.width;
+        let axis = isVertical ? [y, 'y1'] : [x, 'x1'];
 
-        this._dragPlaceholderPos = -1
-        for (let i = children.length - 1; i >= 0; i--) {
+        this._dragPlaceholderPos = -1;
+        let minDist = -1;
+        for(let i = 0; i < children.length; i++) {
             if (!children[i].visible)
                 continue;
-
-            if (isVertical) {
-                if (y > children[i].get_allocation_box().y1) {
-                    this._dragPlaceholderPos = i;
-                    break;
-                }
-            } else if  (x > children[i].get_allocation_box().x1) {
+            let dist = Math.abs(axis[0] - (children[i].get_allocation_box()[axis[1]] + children[i].width / 2));
+            if(dist < minDist || minDist == -1) {
+                minDist = dist;
                 this._dragPlaceholderPos = i;
-                break;
             }
         }
 
@@ -1357,6 +1418,8 @@ class CinnamonWindowListApplet extends Applet.Applet {
 
         this.manager_container.set_child_at_index(source.actor, this._dragPlaceholderPos);
 
+        this._saveOrder();
+
         return true;
     }
 
